<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 1/6/16
 * Time: 8:02 PM
 */

namespace qfq;

use qfq;
use qfq\Store;
use qfq\OnArray;
use qfq\UserFormException;

require_once(__DIR__ . '/../qfq/store/Store.php');
require_once(__DIR__ . '/../qfq/Constants.php');
require_once(__DIR__ . '/../qfq/exceptions/DbException.php');
require_once(__DIR__ . '/../qfq/exceptions/UserFormException.php');
require_once(__DIR__ . '/../qfq/Database.php');
require_once(__DIR__ . '/../qfq/helper/HelperFormElement.php');
require_once(__DIR__ . '/../qfq/helper/Support.php');
require_once(__DIR__ . '/../qfq/helper/OnArray.php');


/**
 * Class AbstractBuildForm
 * @package qfq
 */
abstract class AbstractBuildForm {
    protected $formSpec = array();  // copy of the loaded form
    protected $feSpecAction = array(); // copy of all formElement.class='action' of the loaded form
    protected $feSpecNative = array(); // copy of all formElement.class='native' of the loaded form
    protected $store = null;
    protected $evaluate = null;
    protected $buildElementFunctionName = array();
    protected $pattern = array();
    protected $wrap = array();
    protected $symbol = array();
    protected $showDebugInfo = false;
    protected $inputCheckPattern = array();

//    protected $feDivClass = array(); // Wrap FormElements in <div class="$feDivClass[type]">

    /**
     * @var string
     */
    private $formId = null;

    /**
     * @var Sip
     */
    private $sip = null;

    /**
     * AbstractBuildForm constructor.
     *
     * @param array $formSpec
     * @param array $feSpecAction
     * @param array $feSpecNative
     */
    public function __construct(array $formSpec, array $feSpecAction, array $feSpecNative) {
        $this->formSpec = $formSpec;
        $this->feSpecAction = $feSpecAction;
        $this->feSpecNative = $feSpecNative;
        $this->store = Store::getInstance();
        $this->db = new Database();
        $this->evaluate = new Evaluate($this->store, $this->db);
        $this->showDebugInfo = ($this->store->getVar(SYSTEM_SHOW_DEBUG_INFO, STORE_SYSTEM) === 'yes');

        $this->sip = $this->store->getSipInstance();

        // render mode specific
        $this->fillWrap();

        $this->buildElementFunctionName = [
            'checkbox' => 'Checkbox',
            'date' => 'DateTime',
            'datetime' => 'DateTime',
            'dateJQW' => 'DateJQW',
            'datetimeJQW' => 'DateJQW',
            'email' => 'Input',
            'gridJQW' => 'GridJQW',
            'hidden' => 'Hidden',
            'text' => 'Input',
            'time' => 'DateTime',
            'note' => 'Note',
            'password' => 'Input',
            'radio' => 'Radio',
            'select' => 'Select',
            'subrecord' => 'Subrecord',
            'upload' => 'File',
            'fieldset' => 'Fieldset',
            'pill' => 'Pill'
        ];

        $this->buildRowName = [
            'checkbox' => 'Native',
            'date' => 'Native',
            'datetime' => 'Native',
            'dateJQW' => 'Native',
            'datetimeJQW' => 'Native',
            'email' => 'Native',
            'gridJQW' => 'Native',
            'hidden' => 'Native',
            'text' => 'Native',
            'time' => 'Native',
            'note' => 'Native',
            'password' => 'Native',
            'radio' => 'Native',
            'select' => 'Native',
            'subrecord' => 'Subrecord',
            'upload' => 'Native',
            'fieldset' => 'Fieldset',
            'pill' => 'Pill'
        ];

        $this->symbol[SYMBOL_EDIT] = "<span class='glyphicon glyphicon-pencil'></span>";
        $this->symbol[SYMBOL_NEW] = "<span class='glyphicon glyphicon-plus'></span>";
        $this->symbol[SYMBOL_DELETE] = "<span class='glyphicon glyphicon-trash'></span>";

        $this->inputCheckPattern = Sanitize::inputCheckPatternArray();
    }

    abstract public function fillWrap();

    /**
     * Builds complete form. Depending of form specification, the layout will be 'plain' / 'table' / 'bootstrap'.
     *
     * @param $mode
     * @return string|array   $mode=LOAD_FORM: The whole form as HTML, $mode=FORM_UPDATE: array of all formElement.dynamicUpdate-yes  values/states
     * @throws CodeException
     * @throws DbException
     * @throws \qfq\UserFormException
     */
    public function process($mode, $htmlElementNameIdZero = false) {
        $htmlHead = '';
        $htmlTail = '';
        $htmlSubrecords = '';
        $htmlElements = '';
        $json = array();

        $modeCollectFe = FLAG_DYNAMIC_UPDATE;
        $storeUse = STORE_USE_DEFAULT;

        if ($mode === FORM_SAVE) {
            $modeCollectFe = FLAG_ALL;
            $storeUse = STORE_RECORD . STORE_TABLE_DEFAULT;
        }

        // <form>
        if ($mode === FORM_LOAD) {
            $htmlHead = $this->head();
            $htmlTail = $this->tail();
            $htmlSubrecords = $this->doSubrecords();
        }

        $filter = $this->getProcessFilter();

        if ($this->formSpec['multiMode'] !== 'none') {

            $parentRecords = $this->db->sql($this->formSpec['multiSql']);
            foreach ($parentRecords as $row) {
                $this->store->setVarArray($row, STORE_PARENT_RECORD, true);
                $jsonTmp = array();
                $htmlElements = $this->elements($row['_id'], $filter, 0, $jsonTmp, $modeCollectFe);
                $json[] = $jsonTmp;
            }
        } else {
            $htmlElements = $this->elements($this->store->getVar(SIP_RECORD_ID, STORE_SIP), $filter, 0, $json, $modeCollectFe, $htmlElementNameIdZero, $storeUse);
        }

        $htmlSip = $this->buildHiddenSip($json);

        // </form>

        return ($mode === FORM_LOAD) ? $htmlHead . $htmlElements . $htmlSip . $htmlTail . $htmlSubrecords : $json;
    }

    /**
     * Builds the head area of the form.
     *
     * @return string
     */
    public function head() {
        $html = '';

        $html .= '<div ' . Support::doAttribute('class', $this->formSpec['class'], TRUE) . '>'; // main <div class=...> around everything

        // Logged in BE User will see a FormEdit Link
        $sipParamString = OnArray::toString($this->store->getStore(STORE_SIP), ':', ', ', "'");
        $formEditUrl = $this->createFormEditUrl();

        $html .= "<p><a href='$formEditUrl'>Edit</a> <small>[$sipParamString]</small></p>";

        $html .= $this->wrapItem(WRAP_SETUP_TITLE, $this->formSpec['title'], true);

        $html .= $this->getFormTag();

        return $html;
    }

    /**
     * If SHOW_DEBUG_INFO=yes: create a link (incl. SIP) to edit the current form. Show also the hidden content of the SIP.
     *
     * @return string String: <a href="?pageId&sip=....">Edit</a> <small>[sip:..., r:..., urlparam:..., ...]</small>
     */
    public function createFormEditUrl() {

        if (!$this->showDebugInfo) {
            return '';
        }

        $queryStringArray = [
            'id' => $this->store->getVar(TYPO3_PAGE_ID, STORE_TYPO3),
            'form' => 'form',
            'r' => $this->formSpec['id']
        ];

        $queryString = Support::arrayToQueryString($queryStringArray);

        $sip = $this->store->getSipInstance();
        $url = $sip->queryStringToSip($queryString);

        return $url;
    }

    /**
     * Wrap's $this->wrap[$item][WRAP_SETUP_START] around $value. If $flagOmitEmpty==true && $value=='': return ''.
     *
     * @param $item
     * @param $value
     * @param bool|false $flagOmitEmpty
     * @return string
     */
    public function wrapItem($item, $value, $flagOmitEmpty = false) {
        if ($flagOmitEmpty && $value === "")
            return '';
        return $this->wrap[$item][WRAP_SETUP_START] . $value . $this->wrap[$item][WRAP_SETUP_END];
    }

    /**
     * Returns '<form ...>'-tag with various attributes.
     *
     * @return string
     */
    public function getFormTag() {

        $attribute = $this->getFormTagAtrributes();

        return '<form ' . OnArray::toString($attribute, '=', ' ', "'") . '>';
    }

    /**
     * Build an assoc array with standard form attributes.
     *
     * @return array
     */
    public function getFormTagAtrributes() {

        $attribute['id'] = $this->getFormId();
        $attribute['method'] = 'post';
        $attribute['action'] = $this->getActionUrl();
        $attribute['target'] = '_top';
        $attribute['accept-charset'] = 'UTF-8';
        $attribute['autocomplete'] = 'on';
        $attribute['enctype'] = $this->getEncType();

        return $attribute;
    }

    /**
     * @return string
     */
    public function getFormId() {
        if ($this->formId === null) {
            $this->formId = uniqid('qfq-form-');
        }
        return $this->formId;
    }

    /**
     * Builds the HTML 'form'-tag inlcuding all attributes and target.
     *
     * Notice: the SIP will be transferred as POST Parameter.
     *
     * @return string
     * @throws DbException
     */
    public function getActionUrl() {

        return API_DIR . '/save.php';
    }

    /**
     * Determines the enctype.
     *
     * See: https://www.w3.org/wiki/HTML/Elements/form#HTML_Attributes
     *
     * @return string
     * @throws DbException
     */
    public function getEncType() {

        $result = $this->db->sql("SELECT id FROM FormElement AS fe WHERE fe.formId=? AND fe.type='upload' LIMIT 1", ROW_REGULAR, [$this->formSpec['id']], 'Look for Formelement.type="upload"');
        return (count($result) === 1) ? 'multipart/form-data' : 'application/x-www-form-urlencoded';

    }

    abstract public function tail();

    abstract public function doSubrecords();

    abstract public function getProcessFilter();

    /**
     * Process all FormElements: build corresponding HTML code. Collect and return all HTML code.
     *
     * @param $recordId
     * @param string $filter FORM_ELEMENTS_NATIVE | FORM_ELEMENTS_SUBRECORD | FORM_ELEMENTS_NATIVE_SUBRECORD
     * @param int $feIdContainer
     * @param array $json
     * @param string $modeCollectFe
     * @param bool $htmlElementNameIdZero
     * @param string $storeUse
     * @return string
     * @throws CodeException
     * @throws DbException
     * @throws \qfq\UserFormException
     */
    public function elements($recordId, $filter = FORM_ELEMENTS_NATIVE, $feIdContainer = 0, array &$json,
                             $modeCollectFe = FLAG_DYNAMIC_UPDATE, $htmlElementNameIdZero = false, $storeUse = STORE_USE_DEFAULT) {
        $html = '';

        // get current data record
        if ($recordId > 0 && $this->store->getVar('id', STORE_RECORD) === false) {
            $row = $this->db->sql("SELECT * FROM " . $this->formSpec['tableName'] . " WHERE id = ?", ROW_EXPECT_1, array($recordId));
            $this->store->setVarArray($row, STORE_RECORD);
        }

        // Iterate over all FormElements
        foreach ($this->feSpecNative as $fe) {
            if (($filter === FORM_ELEMENTS_NATIVE && $fe[FE_TYPE] === 'subrecord')
                || ($filter === FORM_ELEMENTS_SUBRECORD && $fe[FE_TYPE] !== 'subrecord')
//                || ($filter === FORM_ELEMENTS_DYNAMIC_UPDATE && $fe['dynamicUpdate'] === 'no')
            ) {
                continue; // skip this FE
            }

            $debugStack = array();

            // Preparation for Log, Debug
            $this->store->setVar(SYSTEM_FORM_ELEMENT, Logger::formatFormElementName($fe), STORE_SYSTEM);

            // evaluate current FormElement
            $evaluate = new Evaluate($this->store, $this->db);
            $formElement = $evaluate->parseArray($fe, $debugStack);

            // Some Defaults
            $formElement = Support::setFeDefaults($formElement);

            Support::setIfNotSet($formElement, F_BS_LABEL_COLUMNS);
            Support::setIfNotSet($formElement, F_BS_INPUT_COLUMNS);
            Support::setIfNotSet($formElement, F_BS_NOTE_COLUMNS);
            $label = ($formElement[F_BS_LABEL_COLUMNS] == '') ? $this->formSpec[F_BS_LABEL_COLUMNS] : $formElement[F_BS_LABEL_COLUMNS];
            $input = ($formElement[F_BS_INPUT_COLUMNS] == '') ? $this->formSpec[F_BS_INPUT_COLUMNS] : $formElement[F_BS_INPUT_COLUMNS];
            $note = ($formElement[F_BS_NOTE_COLUMNS] == '') ? $this->formSpec[F_BS_NOTE_COLUMNS] : $formElement[F_BS_NOTE_COLUMNS];
            $this->fillWrapLabelInputNote($label, $input, $note);

            // Get default value
            $value = ($formElement['value'] === '') ? $this->store->getVar($formElement['name'], $storeUse,
                $formElement['checkType']) : $formElement['value'];

            // Typically: $htmlElementNameIdZero = true
            // After Saving a record, staying on the form, the FormElements on the Client are still known as '<feName>:0'.
            $htmlFormElementId = HelperFormElement::buildFormElementId($formElement['name'], ($htmlElementNameIdZero) ? 0 : $recordId);

            // Construct Marshaller Name: buildElement
            $buildElementFunctionName = 'build' . $this->buildElementFunctionName[$formElement[FE_TYPE]];

            $jsonElement = array();
            // Render pure element
            $elementHtml = $this->$buildElementFunctionName($formElement, $htmlFormElementId, $value, $jsonElement);

//            $fake0 = $fe['dynamicUpdate'];
//            $fake1 = $formElement['dynamicUpdate'];

            // container elements do not have dynamicUpdate='yes'. Instead they deliver nested elements.
            if ($formElement['class'] == 'container') {
                if (count($jsonElement) > 0) {
                    $json = array_merge($json, $jsonElement);
                }
            } else {
                // for non container elements: just add the current json status
                if ($modeCollectFe === FLAG_ALL || ($modeCollectFe == FLAG_DYNAMIC_UPDATE && $fe['dynamicUpdate'] == 'yes')) {
                    if (isset($jsonElement[0]) && is_array($jsonElement[0])) {
                        // Checkboxes are delivered as array of arrays: unnest them and append them to the existing json array.
                        $json = array_merge($json, $jsonElement);
                    } else {
                        $json[] = $jsonElement;
                    }
                }
            }

            // debugStack as Tooltip
            if ($this->showDebugInfo && count($debugStack) > 0) {
//                $elementHtml = Support::appendTooltip($elementHtml, implode("\n", OnArray::htmlentitiesOnArray($debugStack)));
                $elementHtml = Support::appendTooltip($elementHtml, implode("\n", $debugStack));
            }

            // Construct Marshaller Name: buildRow
            $buildRowName = 'buildRow' . $this->buildRowName[$formElement[FE_TYPE]];

            $html .= $this->$buildRowName($formElement, $elementHtml, $htmlFormElementId);
//            break;
        }

        // Log / Debug: Last FormElement has been processed.
        $this->store->setVar(SYSTEM_FORM_ELEMENT, '', STORE_SYSTEM);

        return $html;
    }

    abstract public function fillWrapLabelInputNote($label, $input, $note);

    /**
     * Create a hidden sip, based on latest STORE_SIP Values. Return complete HTML 'hidden' element.
     *
     * @param array $json
     * @return string  <input type='hidden' name='s' value='<sip>'>
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildHiddenSip(array &$json) {
        $sipArray = $this->store->getStore(STORE_SIP);
        unset($sipArray[SIP_SIP]);
        unset($sipArray[SIP_URLPARAM]);

        $queryString = Support::arrayToQueryString($sipArray);
        $sip = $this->store->getSipInstance();

        $sipValue = $sip->queryStringToSip($queryString, RETURN_SIP);

        $json[] = $this->getJsonElementUpdate(CLIENT_SIP, $sipValue, FE_MODE_SHOW);

        return $this->buildNativeHidden(CLIENT_SIP, $sipValue);
    }

    /**
     * @param $htmlFormElementId
     * @param string|array $value
     * @param string $feMode disabled|readonly|''
     * @return array
     */
    private function getJsonElementUpdate($htmlFormElementId, $value, $feMode) {

        $json = $this->getJsonFeMode($feMode);

        $json['form-element'] = $htmlFormElementId;
        $json['value'] = $value;

        return $json;
    }

    /**
     * Set corresponding JSON attributes readonly/required/disabled, based on $formElement[FE_MODE].
     *
     * @param array $feMode
     * @return array
     * @throws UserFormException
     */
    private function getJsonFeMode($feMode) {

        $this->getFeMode($feMode, $hidden, $disabled, $required);

        return [API_JSON_HIDDEN => $hidden === 'yes', API_JSON_DISABLED => $disabled === 'yes', API_JSON_REQUIRED => $required === 'yes'];
    }

    /**
     * @param $feMode
     * @param $hidden
     * @param $disabled
     * @param $required
     * @throws \qfq\UserFormException
     */
    private function getFeMode($feMode, &$hidden, &$disabled, &$required) {
        $hidden = 'no';
        $disabled = 'no';
        $required = 'no';

        switch ($feMode) {
            case FE_MODE_SHOW:
                break;
            case FE_MODE_REQUIRED:
                $required = 'yes';
                break;
            case FE_MODE_READONLY:
                $disabled = 'yes';  // convert the UI status 'readonly' to the HTML/CSS status 'disabled'.
                break;
            case FE_MODE_HIDDEN:
                $hidden = 'yes';
                break;
            default:
                throw new UserFormException("Unknown mode '$feMode'", ERROR_UNKNOWN_MODE);
                break;
        }
    }

    /**
     * Builds a real HTML hidden form element. Useful for checkboxes, Multiple-Select and Radios.
     *
     * @param $htmlFormElementId
     * @param $value
     * @return string
     */
    public function buildNativeHidden($htmlFormElementId, $value) {
        return '<input type="hidden" name="' . $htmlFormElementId . '" value="' . htmlentities($value) . '">';
    }

    /**
     * Takes the current SIP ('form' and additional parameter), set SIP_RECORD_ID=0 and create a new 'NewRecordUrl'.
     *
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function deriveNewRecordUrlFromExistingSip(&$toolTipNew) {
        $urlParam = $this->store->getStore(STORE_SIP);
        $urlParam[SIP_RECORD_ID] = 0;

        unset($urlParam[SIP_SIP]);
        unset($urlParam[SIP_URLPARAM]);

        Support::appendTypo3ParameterToArray($urlParam);

        $sip = $this->store->getSipInstance();

        $url = $sip->queryStringToSip(OnArray::toString($urlParam));

        if ($this->showDebugInfo) {
            //TODO: missing decoding of SIP
            $toolTipNew .= PHP_EOL . PHP_EOL . OnArray::toString($urlParam, ' = ', PHP_EOL, "'");
        }

        return $url;
    }

    abstract public function buildRowNative(array $formElement, $htmlElement, $htmlFormElementId);

    abstract public function buildRowPill(array $formElement, $elementHtml);

    abstract public function buildRowFieldset(array $formElement, $elementHtml);

    abstract public function buildRowSubrecord(array $formElement, $elementHtml);

    /**
     * Builds a label, typically for an html-'<input>'-element.
     *
     * @param string $htmlFormElementId
     * @param string $label
     * @return string
     */
    public function buildLabel($htmlFormElementId, $label) {
        $attributes = Support::doAttribute('for', $htmlFormElementId);
        $attributes .= Support::doAttribute('class', 'control-label');

        $html = Support::wrapTag("<label $attributes>", $label);

        return $html;
    }

    /**
     * Builds HTML 'input' element.
     * Format: <input name="$htmlFormElementId" <type="email|input|password|url" [autocomplete="autocomplete"] [autofocus="autofocus"]
     *           [maxlength="$maxLength"] [placeholder="$placeholder"] [size="$size"] [min="$min"] [max="$max"]
     *           [pattern="$pattern"] [required="required"] [disabled="disabled"] value="$value">
     *
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return string
     * @throws \qfq\UserFormException
     */
    public function buildInput(array $formElement, $htmlFormElementId, $value, array &$json) {
        $textarea = '';

        $attribute = Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('class', 'form-control');

        // Check for input type 'textarea'
        $colsRows = explode(',', $formElement['size'], 2);
        if (count($colsRows) === 2) {
            // <textarea>
            $htmlTag = '<textarea';

            $attribute .= Support::doAttribute('cols', $colsRows[0]);
            $attribute .= Support::doAttribute('rows', $colsRows[1]);
            $textarea = htmlentities($value) . '</textarea>';

        } else {
            $htmlTag = '<input';

            $this->adjustMaxLength($formElement);

            // <input>
            if ($formElement['maxLength'] > 0) {
                // crop string only if it's not empty (substr returns false on empty strings)
                if ($value !== '')
                    $value = substr($value, 0, $formElement['maxLength']);

                // 'maxLength' needs an upper 'L': naming convention for DB tables!
                $attribute .= $this->getAttributeList($formElement, ['type', 'size', 'maxLength']);
                $attribute .= Support::doAttribute('value', htmlentities($value), false);
            }
        }

        $attribute .= $this->getAttributeList($formElement, ['autocomplete', 'autofocus', 'placeholder']);
        $attribute .= Support::doAttribute('data-load', ($formElement['dynamicUpdate'] === 'yes') ? 'data-load' : '');
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= $this->getInputCheckPattern($formElement['checkType'], $formElement['checkPattern']);

        $attribute .= $this->getAttributeFeMode($formElement[FE_MODE]);

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return "$htmlTag $attribute>$textarea" . $this->getHelpBlock();

    }

    /**
     * @param array $formElement
     */
    private function adjustMaxLength(array &$formElement) {

        // MIN( $formElement['maxLength'], tabledefinition)
        $maxLength = $this->getColumnSize($formElement['name']);

        switch ($formElement[FE_TYPE]) {
            case 'date':
                $feMaxLength = 10;
                break;
            case 'datetime':
                $feMaxLength = 19;
                break;
            case 'time':
                $feMaxLength = 8;
                break;
            default:
                $feMaxLength = false;
                break;
        }

        // In case the underlying tablecolumn is not of type date/time, the $maxLength might be to high: correct
        if ($feMaxLength !== false && $maxLength !== false && $feMaxLength < $maxLength) {
            $maxLength = $feMaxLength;
        }

        // date/datetime
        if ($maxLength !== false) {
            if (is_numeric($formElement['maxLength'])) {
                if ($formElement['maxLength'] > $maxLength) {
                    $formElement['maxLength'] = $maxLength;
                }
            } else {
                $formElement['maxLength'] = $maxLength;
            }
        }
    }

    /**
     * Get column spec from tabledefinition and parse size of it. If nothing defined, return false.
     *
     * @param $column
     * @return bool|int  a) 'false' if there is no length definition, b) length definition, c) date|time|datetime|timestamp use hardcoded length
     */
    private function getColumnSize($column) {
        $matches = array();

        $typeSpec = $this->store->getVar($column, STORE_TABLE_COLUMN_TYPES);
        switch ($typeSpec) {
            case 'date': // yyyy-mm-dd
                return 10;
            case 'datetime': // yyyy-mm-dd hh:mm:ss
            case 'timestamp': // yyyy-mm-dd hh:mm:ss
                return 19;
            case 'time': // hh:mm:ss
                return 8;
            default:
                break;
        }

        // e.g.: string(64) >> 64, enum('yes','no') >> false
        if (1 === preg_match('/\((.+)\)/', $typeSpec, $matches)) {
            if (is_numeric($matches[1]))
                return $matches[1];
        }

        return false;
    }

    /**
     * Builds a HTML attribute list, based on  $attributeList.
     *
     * E.g.: attributeList: [ 'type', 'autofocus' ]
     *       generates: 'type="$formElement[FE_TYPE]" autofocus="$formElement['autofocus']" '
     *
     * @param array $formElement
     * @param array $attributeList
     * @return string
     */
    private function getAttributeList(array $formElement, array $attributeList) {
        $attribute = '';
        foreach ($attributeList as $item) {
            if (isset($formElement[$item]))
                $attribute .= Support::doAttribute(strtolower($item), $formElement[$item]);
        }
        return $attribute;
    }

    /**
     * Construct HTML Input attribute for Client Validation:
     *
     *   type     data                      result
     *   -------  -----------------------   -------------------------------------------------------------------------------
     *   min|max  <min value>|<max value>   min="$attrData[0]"|max="$attrData[1]"
     *   pattern  <regexp>                  pattern="$data"
     *   digit    -                         pattern="^[0-9]*$"
     *   email    -                         pattern="^[_a-z0-9-]+(\.[_a-z0-9-]+)*@[a-z0-9-]+(\.[a-z0-9-]+)*(\.[a-z]{2,4})$"
     *   alnumx   -
     *
     * For 'min/max' and 'pattern' the 'data' will be injected in the attribute string via '%s'.
     *
     * @param $type
     * @param $data
     * @return string
     * @throws \qfq\UserFormException
     */
    private function getInputCheckPattern($type, $data) {
        $attribute = '';

        if ($type === '') {
            return '';
        }

        switch ($type) {
            case SANITIZE_ALLOW_MIN_MAX:
            case SANITIZE_ALLOW_MIN_MAX_DATE:
                $arrData = explode("|", $data);
                if (count($arrData) != 2 || $arrData[0] == '' || $arrData[1] == '')
                    throw new UserFormException("Missing MIN|MAX values", ERROR_MISSING_MIN_MAX);

                $attribute = 'min="' . $arrData[0] . '" ';
                $attribute .= 'max="' . $arrData[1] . '" ';
                break;

            case SANITIZE_ALLOW_PATTERN:
                $attribute = 'pattern="' . $data . '" ';
                break;

            case SANITIZE_ALLOW_ALL:
                break;

            default:
                $attribute = 'pattern="' . $this->inputCheckPattern[$type] . '" ';
                break;
        }

        return $attribute;
    }

    /**
     * Set corresponding html attributes readonly/required/disabled, based on $formElement[FE_MODE].
     *
     * @param string $feMode
     * @return string
     * @throws UserFormException
     */
    private function getAttributeFeMode($feMode) {
        $attribute = '';

        $this->getFeMode($feMode, $hidden, $disabled, $required);

        switch ($feMode) {
            case FE_MODE_HIDDEN:
            case FE_MODE_SHOW:
                break;
            case FE_MODE_REQUIRED:
            case FE_MODE_READONLY:
                $attribute .= Support::doAttribute($feMode, $feMode);
                break;
            default:
                throw new UserFormException("Unknown mode '$feMode'", ERROR_UNKNOWN_MODE);
                break;
        }

        // Attributes: data-...
        $attribute .= Support::doAttribute(DATA_HIDDEN, $hidden);
        $attribute .= Support::doAttribute(DATA_DISABLED, $disabled);
        $attribute .= Support::doAttribute(DATA_REQUIRED, $required);

        return $attribute;
    }

    /**
     * @return string
     */
    private function getHelpBlock() {
        return '<div class="help-block with-errors"></div>';
    }

    /**
     * Builds HTML 'checkbox' element.
     *
     * Checkboxes will only be submitted, if they are checked. Therefore, a hidden element with the unchecked value will be transferred first.
     *
     * Format: <input type="hidden" name="$htmlFormElementId" value="$valueUnChecked">
     *         <input name="$htmlFormElementId" type="checkbox" [autofocus="autofocus"]
     *            [required="required"] [disabled="disabled"] value="<value>" [checked="checked"] >
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildCheckbox(array $formElement, $htmlFormElementId, $value, array &$json) {
        $itemKey = array();
        $itemValue = array();

        // Fill $itemKey & $itemValue
        $this->getKeyValueListFromSqlEnumSpec($formElement, $itemKey, $itemValue);

        // Get fallback, if 'checkBoxMode' is not defined:
        if (!isset($formElement['checkBoxMode'])) {
            // This fallback is problematic if 'set' or 'enum' has 2 : defaults to single but maybe multi is meant.
            $formElement['checkBoxMode'] = (count($itemKey) > 2) ? 'multi' : 'single';
        }

        if ($formElement['checkBoxMode'] === 'multi') {
            $htmlFormElementId .= '[]';
        } else {
            // Fill meaningfull defaults to parameter: checked|unchecked  (CHECKBOX_VALUE_CHECKED|CHECKBOX_VALUE_UNCHECKED)
            $this->prepareCheckboxCheckedUncheckedValue($itemKey, $formElement);
        }

        $attributeBase = $this->getAttributeFeMode($formElement[FE_MODE]);
        $attributeBase .= Support::doAttribute('type', $formElement[FE_TYPE]);

        switch ($formElement['checkBoxMode']) {
            case 'single':
                $html = $this->buildCheckboxSingle($formElement, $htmlFormElementId, $attributeBase, $value, $json);
                break;
            case 'multi';
                $html = $this->buildCheckboxMulti($formElement, $htmlFormElementId, $attributeBase, $value, $itemKey, $itemValue, $json);
                break;
            default:
                throw new UserFormException('checkBoxMode: \'' . $formElement['checkBoxMode'] . '\' is unknown.', ERROR_CHECKBOXMODE_UNKNOWN);
        }

        return $html;
    }

    /**
     * Look for key/value list (in this order, first match counts) in
     *  a) `sql1`
     *  b) `parameter:itemList`
     *  c) table.column definition
     *
     * Copies the found keys to &$itemKey and the values to &$itemValue
     * If there are no &$itemKey, copy &$itemValue to &$itemKey.
     *
     * @param array $formElement
     * @param $itemKey
     * @param $itemValue
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function getKeyValueListFromSqlEnumSpec(array $formElement, &$itemKey, &$itemValue) {
        $fieldType = '';
        $itemKey = array();
        $itemValue = array();

        if (count($formElement) < 20)
            throw new CodeException("Invalid (none or to small) Formelement", ERROR_MISSING_FORMELEMENT);

        // Call getItemsForEnumOrSet() only if there a corresponding column really exist.
        if (false !== $this->store->getVar($formElement['name'], STORE_TABLE_COLUMN_TYPES)) {
            $itemValue = $this->getItemsForEnumOrSet($formElement['name'], $fieldType);
        }

        if (is_array($formElement['sql1'])) {
            if (count($formElement['sql1']) > 0) {
                $keys = array_keys($formElement['sql1'][0]);
                $itemKey = array_column($formElement['sql1'], 'id');

                // If there is no column 'id' and at least two columns in total
                if (count($itemKey) === 0 && count($keys) >= 2) {
                    $itemKey = array_column($formElement['sql1'], $keys[0]);
                }

                $itemValue = array_column($formElement['sql1'], 'label');
                // If there is no column 'label' (e.g.: SHOW tables)
                if (count($itemValue) === 0) {
                    $idx = count($keys) == 1 ? 0 : 1;
                    $itemValue = array_column($formElement['sql1'], $keys[$idx]);
                }
            }
        } elseif (isset($formElement['itemList']) && strlen($formElement['itemList']) > 0) {
            $arr = KeyValueStringParser::parse($formElement['itemList'], ':', ',', KVP_IF_VALUE_EMPTY_COPY_KEY);
            $itemValue = array_values($arr);
            $itemKey = array_keys($arr);
        } elseif ($fieldType === 'enum' || $fieldType === 'set') {
            // already done at the beginning with '$this->getItemsForEnumOrSet($formElement['name'], $fieldType);'
        } else {
            throw new UserFormException("Missing definition (- nothing found in 'sql1', 'parameter:itemValues', 'enum-' or 'set-definition'", ERROR_MISSING_ITEM_VALUES);
        }

        if (count($itemKey) === 0) {
            $itemKey = $itemValue;
        }

        if (isset($formElement['emptyItemAtStart'])) {
            array_unshift($itemKey, '');
            array_unshift($itemValue, '');
        }

        if (isset($formElement['emptyItemAtEnd'])) {
            $itemValue[] = '';
            $itemKey[] = '';
        }

        if (isset($formElement['emptyHide'])) {
            if (isset($itemValue['']))
                unset($itemValue['']);
            if (isset($itemKey['']))
                unset($itemKey['']);

        }
    }

    /**
     * Get the attribute definition list of an enum or set column. For strings, get the default value. Return elements as an array.
     *
     * @param $column
     * @param $fieldType
     * @return array
     * @throws UserFormException
     */
    private function getItemsForEnumOrSet($column, &$fieldType) {

        // Get column definition
        $fieldTypeDefinition = $this->store->getVar($column, STORE_TABLE_COLUMN_TYPES);

        if ($fieldTypeDefinition === false) {
            throw new UserFormException("Column '$column' unknown in table '" . $this->formSpec['tableName'] . "'", ERROR_DB_UNKNOWN_COLUMN);
        }

        $length = strlen($fieldTypeDefinition);

        // enum('...   set('
        switch (substr($fieldTypeDefinition, 0, 4)) {
            case 'enum':
                $startPosition = 5;
                break;
            case 'set(':
                $startPosition = 4;
                break;
            default:
                $fieldType = 'string';
                return array();
        }

        // enum('a','b','c', ...)   >> [ 'a', 'b', 'c', ... ]
        // set('a','b','c', ...)   >> [ 'a', 'b', 'c', ... ]
        $items = OnArray::trimArray(explode(',', substr($fieldTypeDefinition, $startPosition, $length - $startPosition - 1)), "'");
        $fieldType = substr($fieldTypeDefinition, 0, $startPosition - 1);

        return $items;
    }

    /**
     * For CheckBox's with only one checkbox: if no parameter:checked|unchecked is defined, take defaults:
     *
     *    checked: first Element in $itemKey
     *  unchecked: ''
     *
     * @param array $itemKey
     * @param array $formElement
     * @throws UserFormException
     */
    private function prepareCheckboxCheckedUncheckedValue(array $itemKey, array &$formElement) {

        if (!isset($formElement[CHECKBOX_VALUE_CHECKED])) {
            if (isset($itemKey[0])) {
                // First element in $itemKey list
                $formElement[CHECKBOX_VALUE_CHECKED] = $itemKey[0];
            } else {
                // Take column default value
                $formElement[CHECKBOX_VALUE_CHECKED] = $this->store->getVar($formElement['name'], STORE_TABLE_DEFAULT);
            }
        }

        // unchecked
        if (!isset($formElement[CHECKBOX_VALUE_UNCHECKED])) {
            if (isset($itemKey[1])) {
                $formElement[CHECKBOX_VALUE_UNCHECKED] = ($itemKey[0] === $formElement['checked']) ? $itemKey[1] : $itemKey[0];
            } else {
                $formElement[CHECKBOX_VALUE_UNCHECKED] = '';
            }
        }

        if ($formElement[CHECKBOX_VALUE_CHECKED] === $formElement[CHECKBOX_VALUE_UNCHECKED]) {
            throw new UserFormException('FormElement: type=checkbox - checked and unchecked can\'t be the same: ' . $formElement[CHECKBOX_VALUE_CHECKED], ERROR_CHECKBOX_EQUAL);
        }

    }

    /**
     * Build a Checkbox based on two values.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $attribute
     * @param $value
     * @param array $json
     * @return string
     */
    public function buildCheckboxSingle(array $formElement, $htmlFormElementId, $attribute, $value, array &$json) {
        $html = '';
        $valueJson = false;

        $attribute .= Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('value', $formElement['checked'], false);
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= Support::doAttribute('data-load', ($formElement['dynamicUpdate'] === 'yes') ? 'data-load' : '');

        if ($formElement['checked'] === $value) {
            $attribute .= Support::doAttribute('checked', 'checked');
            $valueJson = true;
        }

        $attribute .= $this->getAttributeList($formElement, ['autofocus']);

        $html = $this->buildNativeHidden($htmlFormElementId, $formElement['unchecked']);

        $html .= '<input ' . $attribute . '>';
        if (isset($formElement['label2'])) {
            $html .= $formElement['label2'];
        }

        $html = Support::wrapTag("<label>", $html, true);
        $html = Support::wrapTag("<div class='checkbox'>", $html, true);

        $json = $this->getJsonElementUpdate($htmlFormElementId, $valueJson, $formElement[FE_MODE]);

        return $html;
    }

    /**
     * Build as many Checkboxes as items.
     *
     * Layout: The Bootstrap Layout needs very special setup, the checkboxes are wrapped differently with <div class=checkbox>
     *         depending of if they aligned horizontal or vertical.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $attributeBase
     * @param $value
     * @param array $itemKey
     * @param array $itemValue
     * @param array $json
     * @return string
     */
    public function buildCheckboxMulti(array $formElement, $htmlFormElementId, $attributeBase, $value, array $itemKey, array $itemValue, array &$json) {
        $json = array();

        // Defines which of the checkboxes will be checked.
        $values = explode(',', $value);

        $attributeBase .= Support::doAttribute('name', $htmlFormElementId);
        $attributeBase .= Support::doAttribute('data-load', ($formElement['dynamicUpdate'] === 'yes') ? 'data-load' : '');

        $html = $this->buildNativeHidden($htmlFormElementId, '');

        $orientation = ($formElement['maxLength'] > 1) ? ALIGN_HORIZONTAL : ALIGN_VERTICAL;
        $checkboxClass = ($orientation === ALIGN_HORIZONTAL) ? 'checkbox-inline' : 'checkbox';
        $br = '';

        $flagFirst = true;
        for ($ii = 0, $jj = 1; $ii < count($itemKey); $ii++, $jj++) {
            $jsonValue = false;
            $attribute = $attributeBase;

            // Do this only the first round.
            if ($flagFirst) {
                $flagFirst = false;
                if (isset($formElement['autofocus']))
                    $attribute .= Support::doAttribute('autofocus', $formElement['autofocus']);
            }

            $attribute .= Support::doAttribute('value', $itemKey[$ii]);

            // Check if the given key is found in field.
            if (false !== array_search($itemKey[$ii], $values)) {
                $attribute .= Support::doAttribute('checked', 'checked');
                $jsonValue = true;
            }

            // '&nbsp;' - This is necessary to correctly align an empty input.
            $value = ($itemValue[$ii] === '') ? '&nbsp;' : $itemValue[$ii];

            $htmlElement = '<input ' . $attribute . '>' . $value;

//            $htmlElement = Support::wrapTag("<label class='$orientation'>", $htmlElement, true);
//            $htmlElement = Support::wrapTag("<label>", $htmlElement, true);
            // With ALIGN_HORIZONTAL: the label causes some trouble: skip it
            if (($orientation === ALIGN_VERTICAL)) {
                $htmlElement = Support::wrapTag('<label>', $htmlElement);
            }

            $htmlElement = Support::wrapTag("<div class='$checkboxClass'>", $htmlElement, true);

            if ($formElement['maxLength'] > 1) {

                if ($jj == $formElement['maxLength']) {
                    $jj = 0;
                    $br = '<br>';
                } else {
                    $br = '';
                }
            }

            $html .= $htmlElement . $br;
            $json[] = $this->getJsonElementUpdate($htmlFormElementId, $jsonValue, $formElement[FE_MODE]);

        }

//        if (isset($formElement[CHECKBOX_ORIENTATION]) && $formElement[CHECKBOX_ORIENTATION] !== 'vertical')
//            $html = Support::wrapTag("<div class='checkbox'>", $html, true);
//
        return $html;
    }

    /**
     * Submit hidden values by SIP.
     *
     * Sometimes, it's usefull to precalculate values during formload and to submit them as hidden fields.
     * To avoid any manipulation on those fields, the values will be transferred by SIP.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildHidden(array $formElement, $htmlFormElementId, $value, array &$json) {

        $this->store->setVar($htmlFormElementId, $value, STORE_SIP, false);
    }

    /**
     * Build HTML 'radio' element.
     *
     * Checkboxes will only be submitted, if they are checked. Therefore, a hidden element with the unchecked value will be transfered first.
     *
     * Format: <input type="hidden" name="$htmlFormElementId" value="$valueUnChecked">
     *         <input name="$htmlFormElementId" type="radio" [autofocus="autofocus"]
     *            [required="required"] [disabled="disabled"] value="<value>" [checked="checked"] >
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildRadio(array $formElement, $htmlFormElementId, $value, array &$json) {
        $itemKey = array();
        $itemValue = array();

        // Fill $itemKey & $itemValue
        $this->getKeyValueListFromSqlEnumSpec($formElement, $itemKey, $itemValue);

        $attributeBase = $this->getAttributeFeMode($formElement[FE_MODE]);
        $attributeBase .= Support::doAttribute('name', $htmlFormElementId);
        $attributeBase .= Support::doAttribute('type', $formElement[FE_TYPE]);
        $attributeBase .= Support::doAttribute('data-load', ($formElement['dynamicUpdate'] === 'yes') ? 'data-load' : '');

        $jj = 0;
        $flagFirst = true;

        $orientation = ($formElement['maxLength'] > 1) ? ALIGN_HORIZONTAL : ALIGN_VERTICAL;
        $radioClass = ($orientation === ALIGN_HORIZONTAL) ? 'radio-inline' : 'radio';
        $br = '';

        $html = $this->buildNativeHidden($htmlFormElementId, $value);
        for ($ii = 0; $ii < count($itemValue); $ii++) {
            $jj++;
            $attribute = $attributeBase; //
            if ($flagFirst) {
                $flagFirst = false;
                if (isset($formElement['autofocus']))
                    $attribute .= Support::doAttribute('autofocus', $formElement['autofocus']);
            }

            $attribute .= Support::doAttribute('value', $itemKey[$ii]);
            $attribute .= Support::doAttribute('title', $formElement['tooltip']);

            if ($itemKey[$ii] === $value) {
                $attribute .= Support::doAttribute('checked', 'checked');
            }

            // '&nbsp;' - This is necessary to correctly align an empty input.
            $value = ($itemValue[$ii] === '') ? '&nbsp;' : $itemValue[$ii];

            $htmlElement = '<input ' . $attribute . '>' . $value;

            // With ALIGN_HORIZONTAL: the label causes some trouble: skip it
            if (($orientation === ALIGN_VERTICAL)) {
                $htmlElement = Support::wrapTag('<label>', $htmlElement);
            }

            if ($formElement['maxLength'] > 1) {

                if ($jj == $formElement['maxLength']) {
                    $jj = 0;
                    $br = '<br>';
                } else {
                    $br = '';
                }
            }

            $htmlElement = Support::wrapTag("<div class='$radioClass'>", $htmlElement) . $br;

            $html .= $htmlElement;
        }

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return $html;
    }

    /**
     * Builds a Selct (Dropdown) Box.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return mixed
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildSelect(array $formElement, $htmlFormElementId, $value, array &$json) {
        $itemKey = array();
        $itemValue = array();

        // Fill $itemKey & $itemValue
        $this->getKeyValueListFromSqlEnumSpec($formElement, $itemKey, $itemValue);

        $attribute = $this->getAttributeFeMode($formElement[FE_MODE]);
        $attribute .= Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('class', 'form-control');
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= $this->getAttributeList($formElement, ['autofocus']);
        $attribute .= Support::doAttribute('data-load', ($formElement['dynamicUpdate'] === 'yes') ? 'data-load' : '');

        if (isset($formElement['size']) && $formElement['size'] > 1) {
            $attribute .= Support::doAttribute('size', $formElement['size']);
            $attribute .= Support::doAttribute('multiple', 'multiple');
        }

        $option = '';
        $firstSelect = true;
        $jsonValues = array();
        for ($ii = 0; $ii < count($itemValue); $ii++) {
            $option .= '<option ';

            $option .= Support::doAttribute('value', $itemKey[$ii]);

            $jsonValues[] = [
                'value' => $itemKey[$ii],
                'text' => $itemValue[$ii],
                'selected' => ($itemKey[$ii] == $value && $firstSelect)
            ];

            if ($itemKey[$ii] === $value && $firstSelect) {
                $option .= 'selected';
                $firstSelect = false;
            }

            $option .= '>' . $itemValue[$ii] . '</option>';
        }

        $json = $this->getJsonElementUpdate($htmlFormElementId, $jsonValues, $formElement[FE_MODE]);

        return '<select ' . $attribute . '>' . $option . '</select>' . $this->getHelpBlock();
    }

    /**
     * Constuct a HTML table of the subrecord data.
     * Column syntax definition: https://wikiit.math.uzh.ch/it/projekt/qfq/qfq-jqwidgets/Documentation#Type:_subrecord
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildSubrecord(array $formElement, $htmlFormElementId, $value, array &$json) {
        $rcText = false;
        $nameColumnId = 'id';
        $targetTableName = '';
        $flagNew = false;
        $flagEdit = false;
        $flagDelete = false;
        $linkNew = '';
        $control = array();

        $primaryRecord = $this->store->getStore(STORE_RECORD);

        if (!$this->prepareSubrecod($formElement, $primaryRecord, $rcText, $nameColumnId)) {
            return $rcText;
        }

        if (isset($formElement[SUBRECORD_PARAMETER_FORM])) {

            $linkNew = Support::wrapTag('<th>', $this->createFormLink($formElement, 0, $primaryRecord, $this->symbol[SYMBOL_NEW], 'New'));

            // Decode settings in subrecordOption
            $flagNew = Support::findInSet(SUBRECORD_NEW, $formElement['subrecordOption']);
            $flagEdit = Support::findInSet(SUBRECORD_EDIT, $formElement['subrecordOption']);
            if ($flagDelete = Support::findInSet(SUBRECORD_DELETE, $formElement['subrecordOption'])) {
                $targetTableName = $this->getFormTable($formElement[SUBRECORD_PARAMETER_FORM]);
            }
        }

        $columns = $linkNew;

        if (isset($formElement['sql1'][0])) {
            // construct column attributes
            $control = $this->getSubrecordColumnControl(array_keys($formElement['sql1'][0]));

            // Subrecord: Column titles
            $columns .= '<th>' . implode('</th><th>', $control['title']) . '</th>';
        }

        if ($flagDelete)
            $columns .= '<th></th>';

        $html = Support::wrapTag('<tr>', $columns);

        foreach ($formElement['sql1'] as $row) {
            $rowHtml = '';

            if ($flagEdit) {
                $rowHtml .= Support::wrapTag('<td>', $this->createFormLink($formElement, $row[$nameColumnId], $primaryRecord, $this->symbol[SYMBOL_EDIT], 'Edit'));
            } elseif ($flagNew) {
                $rowHtml .= Support::wrapTag('<td>', $rowHtml, false);
            }

            // All columns
            foreach ($row as $columnName => $value) {
                if (isset($control['title'][$columnName]))
                    $rowHtml .= Support::wrapTag('<td>', $this->renderCell($control, $columnName, $value));
            }

            if ($flagDelete) {
                $s = $this->createDeleteUrl($targetTableName, $row['id'], RETURN_SIP);
                $rowHtml .= Support::wrapTag('<td>', Support::wrapTag("<button type='button' class='record-delete' data-sip='$s'>", '<span class="glyphicon glyphicon-trash"></span>'));
            }

            Support::setIfNotSet($row, FE_SUBRECORD_ROW_CLASS);
            $rowClass = 'record ';
            $rowClass .= $row[FE_SUBRECORD_ROW_CLASS];

            Support::setIfNotSet($row, FE_SUBRECORD_ROW_TITLE);
            $rowTitle = $row[FE_SUBRECORD_ROW_TITLE];

            $rowAttribute = Support::doAttribute('class', $rowClass);
            $rowAttribute .= Support::doAttribute('title', $rowTitle);
            $html .= Support::wrapTag("<tr $rowAttribute>", $rowHtml, true);
        }

        return Support::wrapTag('<table class="table table-hover">', $html, true);
    }

    /**
     * Prepare Subrecord:
     * - check if the current record has an recordId>0. If not, subrecord can't be edited. Return a message.
     * - check if there is an SELECT statement for the subrecords.
     *
     * @param $formElement
     * @param $primaryRecord
     * @param $rcText
     * @param $nameColumnId
     * @return bool
     * @throws \qfq\UserFormException
     */
    private function prepareSubrecod(array $formElement, array $primaryRecord, &$rcText, &$nameColumnId) {

        if (!isset($primaryRecord['id'])) {
            $rcText = 'Please save record first.';
            return false;
        }

        if (!is_array($formElement['sql1'])) {
            throw new UserFormException('Missing \'sql1\' Query', ERROR_MISSING_SQL1);
        }

        // No records?
        if (count($formElement['sql1']) == 0) {
            $rcText = '';
            return true;
        }

        if (!isset($formElement['sql1'][0][$nameColumnId]))
            $nameColumnId = '_id';

        if (!isset($formElement['sql1'][0][$nameColumnId])) {
            throw new UserFormException('Missing column \'id\' (or "_id") in  \'sql1\' Query', ERROR_DB_MISSING_COLUMN_ID);
        }

        return true;
    }

    /**
     * Renders an Link with a symbol (edit/new) and register a new SIP to grant permission to the link..
     *
     * Returns <a href="<Link>">[icon]</a>
     *
     * Link: <page>?s=<SIP>&<standard typo3 params>
     * SIP: form = $formElement['form'] (provided via formElement['parameter'])
     *      r = $targetRecordId
     *      Parse  $formElement['detail'] with possible key/value pairs. E.g.: detail=id:gr_id,#{{a}}:p_id,#12:x_id
     *        gr_id = <<primarytable.id>>
     *        p_id = <<variable defined in SIP or Client>>
     *        x_id = 12 (constant)
     *
     *
     * @param $formElement
     * @param $targetRecordId
     * @param $record
     * @return string
     * @throws UserFormException
     */
    private function createFormLink(array $formElement, $targetRecordId, array $record, $symbol, $toolTip) {

        $queryStringArray = [
            SIP_FORM => $formElement[SUBRECORD_PARAMETER_FORM],
            SIP_RECORD_ID => $targetRecordId
        ];

        // Add custom query parameter
        if (isset($formElement[SUBRECORD_PARAMETER_DETAIL])) {
            $detailParam = KeyValueStringParser::parse($formElement[SUBRECORD_PARAMETER_DETAIL]);
            foreach ($detailParam as $src => $dest) {
                // Constants
                if ($src[0] == '&') {
                    $queryStringArray[$dest] = substr($src, 1);
                    continue;
                }
                // Form record values or parameter
                if (isset($record[$src])) {
                    $queryStringArray[$dest] = $record[$src];
                }
            }
        }

        if ($this->showDebugInfo)
            $toolTip .= PHP_EOL . OnArray::toString($queryStringArray, ' = ', PHP_EOL, "'");

        Support::appendTypo3ParameterToArray($queryStringArray);
        // If there is a specific targetpage defined, take it.
        if (isset($formElement[SUBRECORD_PARAMETER_PAGE]) && $formElement[SUBRECORD_PARAMETER_PAGE] !== '') {
            $queryStringArray['id'] = $formElement[SUBRECORD_PARAMETER_PAGE];
        }

        //-----------------
        $queryString = Support::arrayToQueryString($queryStringArray);

        $sip = $this->store->getSipInstance();
        $url = $sip->queryStringToSip($queryString);

        return Support::wrapTag('<a class="btn btn-default" href="' . $url . '" title="' . $toolTip . '">', $symbol);
    }

    /**
     * Get the name for the given form $formName. If not found, return ''.
     *
     * @param $formName
     * @return string   tableName for $formName
     * @throws CodeException
     * @throws DbException
     */
    private function getFormTable($formName) {
        $row = $this->db->sql("SELECT tableName FROM Form AS f WHERE f.name = ?", ROW_EXPECT_0_1, [$formName]);
        if (isset($row['tableName'])) {
            return $row['tableName'];
        }

        return '';
    }

    /**
     * Get various column format information based on the 'raw' column title. The attributes are separated by '|' and specified as 'key' or 'key=value'.
     *
     * - Return all parsed values as an assoc array.
     * - For regular columns: If there is no 'width' specified, take the default 'SUBRECORD_COLUMN_WIDTH'
     * - For 'icon /  url / mailto': no width limit.
     *
     * Returned assoc array:
     *  title      Only key. Element is non numeric, which is not a keyword 'width/nostrip/icon/url/mailto'
     *  width      Key/Value Pair. Not provided for 'icon/url/mailto'.
     *  nostrip    Only key. Do not strip HTML Tags from the content.
     *  icon       Only key. Value will rendered (later) as an image.
     *  url        Only key. Value will rendered (later) as a 'href'
     *  mailto     Only key. Value will rendered (later) as a 'href mailto'
     *
     *
     * @param $titleRaw
     * @return array
     * @throws UserFormException
     */
    private function getSubrecordColumnControl(array $titleRaw) {
        $control = array();

        foreach ($titleRaw AS $columnName) {

            switch ($columnName) {
                case FE_SUBRECORD_ROW_CLASS:
                case FE_SUBRECORD_ROW_TITLE:
                    continue 2;
                default:
                    break;
            }

            $flagWidthLimit = true;
            $control['width'][$columnName] = SUBRECORD_COLUMN_WIDTH;

            // a) 'City@width=40', b) 'Status@icon', c) 'Mailto@width=80@nostrip'
            $arr = KeyValueStringParser::parse($columnName, '=', '|', KVP_IF_VALUE_EMPTY_COPY_KEY);
            foreach ($arr as $attribute => $value) {
                switch ($attribute) {
                    case 'width':
                    case 'nostrip':
                    case 'title':
                        break;
                    case 'icon':
                    case 'url':
                    case 'mailto':
                        $flagWidthLimit = false;
                        break;
                    default:
                        $attribute = is_numeric($value) ? 'width' : 'title';
                        break;
                }
                $control[$attribute][$columnName] = $value;
            }

            if (!isset($control['title'][$columnName]))
                $control['title'][$columnName] = ''; // Fallback:  Might be wrong, but better than nothing.

            // Limit title length
            $control['title'][$columnName] = substr($control['title'][$columnName], 0, $control['width'][$columnName]);

            if (!$flagWidthLimit) {
                $control['width'][$columnName] = false;
            }

        }
        return $control;
    }

    /**
     * Renders $value as specified in array $control
     *
     * nostrip: by default, HTML tags are removed. With this attribute, the value will be delivered as it is.
     * width: if there is a size limit - apply it.
     * icon: The cell will be rendered as an image. $value should contain the name of an image in 'fileadmin/icons/'
     * mailto: The cell will be rendered as an <a> tag with the 'mailto' attribute.
     * url:  The cell will be rendered as an <a> tag. The value will be exploded by '|'. $value[0] = href, value[1] = text.
     *  E.g. $value = 'www.math.uzh.ch/?id=45&v=234|Show details for Modul 123' >> <a href="www.math.uzh.ch/?id=45&v=234">Show details for Modul 123</a>
     *
     * @param array $control
     * @param $columnName
     * @param $value
     * @return string
     */
    private function renderCell(array $control, $columnName, $value) {

        switch ($columnName) {
            case FE_SUBRECORD_ROW_CLASS:
            case FE_SUBRECORD_ROW_TITLE:
                return '';
            default:
                break;
        }

        $arr = explode('|', $value);
        if (count($arr) == 1)
            $arr[1] = $arr[0];

        $cell = isset($control['nostrip'][$columnName]) ? $value : strip_tags($value);

        if ($control['width'][$columnName] !== false)
            $cell = substr($cell, 0, $control['width'][$columnName]);

        if (isset($control['icon'][$columnName])) {
            $cell = ($cell === '') ? '' : "<image src='fileadmin/icons/$cell'>";
        }

        if (isset($control['mailto'][$columnName])) {
            $cell = "<a href='mailto:$arr[0]'>$arr[1]</a>";
        }

        if (isset($control['url'][$columnName])) {
            $cell = "<a href='$arr[0]'>$arr[1]</a>";
        }

        return $cell;
    }

    /**
     * Create a link (incl. SIP) to delete the current record.
     *
     * @return string String: "API_DIR/delete.php?sip=...."
     */
    public function createDeleteUrl($table, $recordId, $mode = RETURN_URL) {

        $queryStringArray = [
            SIP_TABLE => $table,
            SIP_RECORD_ID => $recordId
        ];

        $queryString = Support::arrayToQueryString($queryStringArray);

        $sip = $this->store->getSipInstance();

        return $sip->queryStringToSip($queryString, $mode, API_DIR . '/delete.php');
    }

    /**
     * Builds an Upload (File) Button.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return string
     * @throws CodeException
     * @throws \qfq\UserFormException
     */
    public function buildFile(array $formElement, $htmlFormElementId, $value, array &$json) {
        $attribute = '';

        $arr = array();
        $arr['fake_uniq_never_use_this'] = uniqid(); // make sure we get a new SIP. This is needed for multiple forms (same user) with r=0
        $arr[CLIENT_SIP_FOR_FORM] = $this->store->getVar(SIP_SIP, STORE_SIP);
        $arr[CLIENT_FE_NAME] = $formElement['name'];
        $arr[CLIENT_FORM] = $this->formSpec['name'];
        $arr[CLIENT_RECORD_ID] = $this->store->getVar(SIP_RECORD_ID, STORE_SIP);
        $arr[CLIENT_PAGE_ID] = 'fake';
        $sipUpload = $this->sip->queryStringToSip(OnArray::toString($arr), RETURN_SIP);

        $hiddenSipUpload = $this->buildNativeHidden($htmlFormElementId, $sipUpload);

        $attribute .= Support::doAttribute('name', $htmlFormElementId);
//        $attribute .= Support::doAttribute('class', 'form-control');
        $attribute .= Support::doAttribute('type', 'file');
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= $this->getAttributeList($formElement, ['autofocus', 'accept']);
        $attribute .= Support::doAttribute('data-load', ($formElement['dynamicUpdate'] === 'yes') ? 'data-load' : '');
        $attribute .= Support::doAttribute('data-sip', $sipUpload);

        if ($value === '') {
            $textDeleteClass = 'hidden';
            $uploadClass = '';
        } else {
            $textDeleteClass = '';
            $uploadClass = 'hidden';
            $formElement[FE_MODE] = FE_MODE_HIDDEN;
        }

        $attribute .= $this->getAttributeFeMode($formElement[FE_MODE]);
        $attribute .= Support::doAttribute('class', $uploadClass, true);
        $htmlInputFile = '<input ' . $attribute . '>' . $this->getHelpBlock();

        $deleteButton = Support::wrapTag("<button class='delete-file' data-sip='$sipUpload' name='delete-$htmlFormElementId'>", $this->symbol[SYMBOL_DELETE]);

        $htmlFilename = Support::wrapTag("<span class='uploaded-file-name'>", $value, false);
        $htmlTextDelete = Support::wrapTag("<div class='uploaded-file $textDeleteClass'>", $htmlFilename . ' ' . $deleteButton);

//        <button type="button" class="file-delete" data-sip="571d1fc9e6974"><span class="glyphicon glyphicon-trash"></span></button>

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return $htmlTextDelete . $htmlInputFile . $hiddenSipUpload;
    }

    /**
     * Builds HTML 'input' element.
     * Format: <input name="$htmlFormElementId" <type="email|input|password|url" [autocomplete="autocomplete"] [autofocus="autofocus"]
     *           [maxlength="$maxLength"] [placeholder="$placeholder"] [size="$size"] [min="$min"] [max="$max"]
     *           [pattern="$pattern"] [required="required"] [disabled="disabled"] value="$value">
     *
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return string
     * @throws \qfq\UserFormException
     */
    public function builddateJQW(array $formElement, $htmlFormElementId, $value, array &$json) {
        $textarea = '';

        $attribute = Support::doAttribute('name', $htmlFormElementId);

        $htmlTag = '<input';

        $this->adjustMaxLength($formElement);

        // Get date format
        if (!isset($formElement['dateFormat'])) {
            $formElement['dateFormat'] = $this->store->getVar(SYSTEM_DATE_FORMAT, STORE_SYSTEM);
        }
        // Format date
        $value = Support::convertDateTime($value, $formElement[FE_DATE_FORMAT], $formElement[FE_SHOW_ZERO], 0, $formElement[FE_SHOW_SECONDS]);

        // <input>
        if ($formElement['maxLength'] > 0) {
            // crop string only if it's not empty (substr returns false on empty strings)
            if ($value !== '')
                $value = substr($value, 0, $formElement['maxLength']);

            // 'maxLength' needs an upper 'L': naming convention for DB tables!
            $attribute .= $this->getAttributeList($formElement, ['type', 'size', 'maxLength']);
            $attribute .= Support::doAttribute('value', htmlentities($value), false);
        }

        $attribute .= $this->getAttributeList($formElement, ['autocomplete', 'autofocus', 'placeholder']);
        $attribute .= Support::doAttribute('data-load', ($formElement['dynamicUpdate'] === 'yes') ? 'data-load' : '');
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= $this->getInputCheckPattern($formElement['checkType'], $formElement['checkPattern']);

        $attribute .= $this->getAttributeFeMode($formElement[FE_MODE]);

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return "$htmlTag $attribute>$textarea";

    }

    /**
     * Builds HTML 'input' element.
     * Format: <input name="$htmlFormElementId" <type="date" [autocomplete="autocomplete"] [autofocus="autofocus"]
     *           [maxlength="$maxLength"] [placeholder="$placeholder"] [size="$size"] [min="$min"] [max="$max"]
     *           [pattern="$pattern"] [required="required"] [disabled="disabled"] value="$value">
     *
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return string
     * @throws UserFormException
     */
    public function buildDateTime(array $formElement, $htmlFormElementId, $value, array &$json) {

        $attribute = Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('class', 'form-control');

        $arrMinMax = null;

        $this->adjustMaxLength($formElement);
        $showTime = ($formElement[FE_TYPE] == 'time' || $formElement[FE_TYPE] == 'datetime') ? 1 : 0;
        $value = Support::convertDateTime($value, $formElement[FE_DATE_FORMAT], $formElement[FE_SHOW_ZERO], $showTime, $formElement[FE_SHOW_SECONDS]);

        $tmpPattern = $formElement['checkPattern'];
        $formElement['checkPattern'] = Support::dateTimeRegexp($formElement[FE_TYPE], $formElement['dateFormat']);


        switch ($formElement['checkType']) {
            case  SANITIZE_ALLOW_PATTERN:
                $formElement['checkPattern'] = $tmpPattern;
                break;
            case SANITIZE_ALLOW_MIN_MAX_DATE:
                $arrMinMax = explode('|', $formElement['checkPattern'], 2);
                if (count($arrMinMax) != 2) {
                    throw new UserFormException('Missing min|max definition', ERROR_MISSING_MIN_MAX);
                }
                break;
            case SANITIZE_ALLOW_ALL:
            case SANITIZE_ALLOW_ALNUMX:
            case SANITIZE_ALLOW_ALLBUT:
                $formElement['checkType'] = SANITIZE_ALLOW_PATTERN;
                break;
            default:
                throw new UserFormException("Checktype not applicable for date/time: '" . $formElement['checkType'] . "'", ERROR_NOT_APPLICABLE);
        }

        // truncate if necessary
        if ($value != '' && $formElement['maxLength'] > 0) {
            $value = substr($value, 0, $formElement['maxLength']);
        }

        $type = $formElement[FE_TYPE];
        if ($type === 'datetime')
            $type = 'datetime-local';

        $attribute .= Support::doAttribute('type', $type);


        // 'maxLength' needs an upper 'L': naming convention for DB tables!
        $attribute .= $this->getAttributeList($formElement, ['size', 'maxLength']);
        $attribute .= Support::doAttribute('value', htmlentities($value), false);

        if ($formElement['placeholder'] == '') {
            $timePattern = ($formElement[FE_SHOW_SECONDS] == 1) ? 'hh:mm:ss' : 'hh:mm';
            switch ($formElement[FE_TYPE]) {
                case 'date':
                    $placeholder = $formElement['dateFormat'];
                    break;
                case 'datetime':
                    $placeholder = $formElement['dateFormat'] . ' ' . $timePattern;
                    break;
                case 'time':
                    $placeholder = $timePattern;
                    break;
                default:
                    throw new UserFormException("Unexpected Formelement type: '" . $formElement[FE_TYPE] . "'", ERROR_FORMELEMENT_TYPE);
            }
            $formElement['placeholder'] = $placeholder;
        }

        $attribute .= $this->getAttributeList($formElement, ['autocomplete', 'autofocus', 'placeholder']);
        $attribute .= Support::doAttribute('data-load', ($formElement['dynamicUpdate'] === 'yes') ? 'data-load' : '');
        $attribute .= Support::doAttribute('title', $formElement['tooltip']);
        $attribute .= $this->getInputCheckPattern($formElement['checkType'], $formElement['checkPattern']);

        if (is_array($arrMinMax)) {
            $attribute .= Support::doAttribute('min', $arrMinMax[0]);
            $attribute .= Support::doAttribute('max', $arrMinMax[1]);
        }

        $attribute .= $this->getAttributeFeMode($formElement[FE_MODE]);

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return "<input $attribute>" . $this->getHelpBlock();

    }

    /**
     * Build Grid JQW element.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @throws UserFormException
     */
    public function buildGridJQW(array $formElement, $htmlFormElementId, $value) {
        // TODO: implement
        throw new UserFormException("Not implemented yet: buildGridJQW()", ERROR_NOT_IMPLEMENTED);
    }

    /**
     * Build Note.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return mixed
     */
    public function buildNote(array $formElement, $htmlFormElementId, $value, array &$json) {
        return $value;
    }

    /**
     * Build Pill
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return mixed
     */
    public function buildPill(array $formElement, $htmlFormElementId, $value, array &$json) {
        return $value;
    }

    /**
     * Build a HTML fieldset. Renders all assigned FormElements inside the fieldset.
     *
     * @param array $formElement
     * @param $htmlFormElementId
     * @param $value
     * @param array $json
     * @return mixed
     * @throws CodeException
     * @throws DbException
     */
    public function buildFieldset(array $formElement, $htmlFormElementId, $value, array &$json) {
        $attribute = '';

        // save parent processed FE's
        $tmpStore = $this->feSpecNative;

        $attribute .= Support::doAttribute('name', $htmlFormElementId);
        $attribute .= Support::doAttribute('data-load', ($formElement['dynamicUpdate'] === 'yes') ? 'data-load' : '');

        // <fieldset>
        $html = '<fieldset ' . $attribute . '>';


        if ($formElement['label'] !== '') {
            $html .= '<legend>' . $formElement['label'] . '</legend>';
        }

        $html .= $this->wrap[WRAP_SETUP_IN_FIELDSET][WRAP_SETUP_START];

        // child FE's
        $sql = SQL_FORM_ELEMENT_SPECIFIC_CONTAINER;
        $this->feSpecNative = $this->db->sql($sql, ROW_REGULAR, ['yes', $this->formSpec["id"], 'native,container', $formElement['id']]);
        HelperFormElement::explodeParameterInArrayElements($this->feSpecNative);
        $html .= $this->elements($this->store->getVar(SIP_RECORD_ID, STORE_SIP), FORM_ELEMENTS_NATIVE_SUBRECORD, 0, $json);


        $html .= $this->wrap[WRAP_SETUP_IN_FIELDSET][WRAP_SETUP_END];

        $html .= '</fieldset>';

        // restore parent processed FE's
        $this->feSpecNative = $tmpStore;

        $json = $this->getJsonElementUpdate($htmlFormElementId, $value, $formElement[FE_MODE]);

        return $html;
    }

    /**
     * @param $table
     * @param $recordId
     * @param $symbol
     * @param $toolTip
     * @return string
     */
    private function createDeleteLink($table, $recordId, $symbol, $toolTip) {

        if ($this->showDebugInfo) {
            $toolTip .= PHP_EOL . "table = '$table'" . PHP_EOL . "id = '$recordId'";
        }

        $url = $this->createDeleteUrl($table, $recordId);

        return Support::wrapTag('<a href="' . $url . '" title="' . $toolTip . '">', $symbol);

    }

}